module Lemmas where

open import Data.Bool hiding (_≤_)
open import Data.Nat
open import Data.Nat.Properties
open import Data.Maybe
open import Relation.Binary.PropositionalEquality

postulate
  funext : ∀ {l₁ l₂} → {A : Set l₁}{B : A → Set l₂} → {f g : (x : A) → B x} → ((x : A) → f x ≡ g x) → f ≡ g

+-over-≤ : {m n o : ℕ} → n ≤ o → m + n ≤ m + o
+-over-≤ {zero} {n} {o} x = x
+-over-≤ {(suc n₁)} {n} {o} x = s≤s (+-over-≤ x)

s≤z-absurd : ∀ {l} {Whatever : Set l}{n : ℕ} → .(suc n ≤ zero) → Whatever
s≤z-absurd ()

m+sn≤i : {m n i : ℕ} → m + suc n ≤ i → suc (m + n) ≤ i
m+sn≤i {m} {n} {i} p rewrite +-suc m n = p

m+s≤z-absurd : ∀ {l} {Whatever : Set l}{m n : ℕ} → .(m + suc n ≤ zero) → Whatever
m+s≤z-absurd p = s≤z-absurd (m+sn≤i p)

sabc≡asbc : {a b c : ℕ} → a + suc (b + c) ≡ suc (a + b + c)
sabc≡asbc {zero} {b} {c} = refl
sabc≡asbc {suc a} {b} {c} = cong suc sabc≡asbc

x+sy : (x y z : ℕ) → (suc (suc (x + y)) ≤ suc z) → x + suc y ≤ z
x+sy x y z prf rewrite (+-suc x y) = ≤-pred prf

n+ssm≤si : {n m i : ℕ} → n + suc (suc m) ≤ suc i → n + suc m ≤ i
n+ssm≤si p = ≤-pred (m+sn≤i p)

cong₂-irr : ∀ {a b c} {A : Set a} {B : A → Set b} {C : Set c}
        (f : (a : A) → .(B a) → C) {x y}.{u v} → x ≡ y → f x u ≡ f y v
cong₂-irr f refl = refl

maybeDistributes : ∀ {l₁ l₂ l₃}{A : Set l₁}{B : Set l₂}{C : Set l₃} →
  (f : B → C)(a : A → B)(b : B)(m : Maybe A) →
  f (maybe′ (λ n → a n) b m) ≡ maybe′ (λ n → f (a n)) (f b) m
maybeDistributes f a b (just x) = refl
maybeDistributes f a b nothing = refl

maybeFlatten₃ : {A B C D : Set}
  → (d : D)(x : Maybe A)(y : Maybe B)(f : C → D)(g : A → B → C)
  → maybe′ f d (maybe′ (λ n → maybe′ (λ m → just (g n m)) nothing y) nothing x)
  ≡
    maybe′ (λ n → maybe′ (λ m → f (g n m)) d y) d x
maybeFlatten₃ d (just x) (just y) f g = refl
maybeFlatten₃ d (just x) nothing f g = refl
maybeFlatten₃ d nothing y f g = refl

maybeFlatten₂ : {A B : Set}
  (f : A → B)(b : B)(x : Maybe A)(y : Maybe A)
  → maybe′ f b (maybe′ just y x)
  ≡
    maybe′ f (maybe′ f b y) x
maybeFlatten₂ f b (just x) y = refl
maybeFlatten₂ f b nothing y = refl

if-distributes : ∀ {l₁ l₂} {A : Set l₁}{B : Set l₂}(f : A → B)(b : Bool)(x y : A)
  → f (if b then x else y) ≡ (if b then f x else f y)
if-distributes f false x y = refl
if-distributes f true x y = refl
